unit DsFancyButton;

{-----------------------------------------------
 DsFancyButton Release 2.0 for Delphi 3/4/5
 Copyright (c) 1999-2001 by Djoko Susilo
 e-mail: djokos@cabi.net.id
 ----------- Q14 Project Feb-99 R1.0 -----------
 Files needed to install:
 DsFancyButton.pas (this file)
 DsFancyButton.dcr (resource for icon)
 -----------------------------------------------
 R1.1 10-Nov-99:
 -Gradient process using region function
 R1.1 22-May-00:
 -Using better region routine to release memory
 R2.0 10-Jan-01:
 -Shape constant of shRoundRect is removed.
  Rounded rectangle can be set through CornerRadius setting.
 -Add property glyph and its layout.
  Max. glyph number is 3 (0=enable  1=disable  2=pressed)
 -Better button pressing effect.
 -----------------------------------------------
 thanks to:
   - Danny Indra <dani@danindra.com>
   - Parker Nolen <nolenj@worldnet.att.net>
 for feedback and suggestions.
 -----------------------------------------------}


{$R-}
{$B-}
{$H+}   {--always use this compiler directive----}

{--for Delphi 5 only--}
//{$DEFINE Delphi5} {--disable this line for previous Delphi version--}

interface

uses
  SysUtils, Windows, Messages, Classes, Graphics, Controls, Forms;

type
  TBmpLayout = (lyLeft, lyTop, lyRight, lyBottom, lyCenter);
  TTextStyle = (txNone, txLowered, txRaised, txShadowed);
  TShape = (shCapsule, shOval, shRectangle);
  TDsFancyButton = class(TGraphicControl)
  private
    FButtonColor: TColor;
    FIsDown: Boolean;
    FFrameColor: TColor;
    FFrameWidth: Integer;
    FCornerRadius: Integer;
    MRgn: HRgn;
    FShape: TShape;
    FTextColor: TColor;
    FTextStyle: TTextStyle;
    FBmpGlyph: TBitmap;
    FGlyphsNum: Byte;
    FGlyphSpace: Integer;
    FBmpLayout: TBmpLayout;
    FBmpWidth: Integer;
    FTransparentColor: TColor;

    procedure SetButtonColor(Value: TColor);
    procedure CMEnabledChanged(var message: TMessage);
              message CM_ENABLEDCHANGED;
    procedure CMTextChanged(var message: TMessage);
              message CM_TEXTCHANGED;
    procedure CMDialogChar(var message: TCMDialogChar);
              message CM_DIALOGCHAR;
    procedure WMSize(var message: TWMSize); message WM_PAINT;
  protected
    procedure Click; override;
    procedure DrawShape;
    procedure DrawGlyph;
    procedure LayoutSetting;
    procedure Paint; override;
    procedure SetFrameColor(Value: TColor);
    procedure SetFrameWidth(Value: Integer);
    procedure SetCornerRadius(Value: Integer);
    procedure SetShape(Value: TShape);
    procedure SetTextStyle(Value: TTextStyle);
    procedure SetBmpLayout(Value: TBmpLayout);
    procedure SetBmpGlyph(ABmpGlyph: TBitmap);
    procedure SetGlyphsNum(Value: Byte);
    procedure SetGlyphSpace(Value: Integer);
    procedure SetTransparentColor(NewColor: TColor);
    procedure MouseUp(Button: TMouseButton;
              Shift: TShiftState; X, Y: Integer); override;
    procedure MouseDown(Button: TMouseButton;
              Shift: TShiftState; X, Y: Integer); override;
    procedure WriteCaption;
    procedure DrawButtonRegion;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  published
    {$IFDEF Delphi5}
    property Anchors;
    property Constraints;
    {$ENDIF}
    property ButtonColor: TColor
             read FButtonColor write SetButtonColor;
    property Caption;
    property DragCursor;
    property DragMode;
    property Enabled;
    property Font;
    property FrameColor: TColor
             read FFrameColor write SetFrameColor;
    property FrameWidth: Integer
             read FFrameWidth write SetFrameWidth;
    property Glyph: TBitmap read FBmpGlyph write SetBmpGlyph;
    property GlyphLayout: TBmpLayout
             read FBmpLayout write SetBmpLayout;
    property GlyphsNum: Byte read FGlyphsNum write SetGlyphsNum;
    property GlyphSpace: Integer read FGlyphSpace write SetGlyphSpace;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property CornerRadius: Integer
             read FCornerRadius write SetCornerRadius;
    property Shape: TShape
             read FShape write SetShape;
    property ShowHint;
    property TextStyle: TTextStyle
             read FTextStyle write SetTextStyle;
    property TransparentColor: TColor
             read FTransparentColor write SetTransparentColor;
    property Visible;

    property OnClick;         property OnDragDrop;
    property OnDragOver;      property OnEndDrag;
    property OnMouseDown;     property OnMouseUp;
    property OnMouseMove;
  end;

procedure Register;

implementation

var GlyphL, GlyphT: Integer;
    TextL, TextT: Integer;
    IsInRegion: Boolean;

constructor TDsFancyButton.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  ControlStyle:=[csClickEvents, csCaptureMouse, csSetCaption];
  Enabled:=True;
  FBmpGlyph:=TBitmap.Create;
  FBmpLayout:=lyLeft;
  FBmpWidth:=0;
  FGlyphsNum:=1;
  FGlyphSpace:=0;
  FButtonColor:=clBtnFace;
  FIsDown:=False;
  FFrameColor:=clGray;
  FFrameWidth:=6;
  FCornerRadius:=10;
  FShape:=shRectangle;
  FTextStyle:=txRaised;
  Height:=25;
  Visible:=True;
  Width:=97;
end;

destructor TDsFancyButton.Destroy;
begin
  if Assigned(FBmpGlyph) then FBmpGlyph.Destroy;
  inherited Destroy;
end;

procedure TDsFancyButton.Paint;
var ClrUp, ClrDown: TColor;
begin
  FBmpWidth:=FBmpGlyph.Width div FGlyphsNum;

  Canvas.Brush.Style:=bsClear;

  if FIsDown then
    begin ClrUp:=clBlack; ClrDown:=clBtnHighlight; end
  else
    begin ClrUp:=clBtnShadow; ClrDown:=clBtnShadow; end;

  DrawButtonRegion;
  try
    Canvas.Brush.Color:=FButtonColor;
    FillRgn(Canvas.Handle, MRgn, Canvas.Brush.Handle);
    Canvas.Brush.Color:=ClrUp;
    FrameRgn(Canvas.Handle, MRgn, Canvas.Brush.Handle, 1, 1);
    OffsetRgn(MRgn, 1, 1);
    Canvas.Brush.Color:=ClrDown;
    FrameRgn(Canvas.Handle, MRgn, Canvas.Brush.Handle, 1, 1);
  finally
    DeleteObject(MRgn);
  end;
  DrawShape;
  LayoutSetting;
  if FBmpWidth>0 then DrawGlyph;
  WriteCaption;
end;

procedure TDsFancyButton.DrawButtonRegion;
var Dia: Integer;
begin
  case Shape of
    shRectangle:
      begin
        Dia:=2*CornerRadius;
        if Dia=0 then MRgn:=CreateRectRgn(0, 0, Width-1, Height-1)
        else MRgn:=CreateRoundRectRgn(0, 0, Width, Height, Dia, Dia);
      end;
    shCapsule:
      begin
        if Width<Height then Dia:=Width else Dia:=Height;
        MRgn:=CreateRoundRectRgn(0, 0, Width, Height, Dia, Dia);
      end;
    shOval: MRgn:=CreateEllipticRgn(0, 0, Width, Height);
  end; //case
end;

procedure TDsFancyButton.DrawShape;
var
  FC, Warna: TColor;
  R, G, B: Byte;
  AwalR, AwalG, AwalB,
  AkhirR, AkhirG, AkhirB,
  n, t, Dia: Integer;
  FRgn: HRgn;
  TempValue: real;
begin
  if FFrameWidth mod 2=0 then t:=FFrameWidth
  else t:=FFrameWidth+1;

  Warna:=ColorToRGB(ButtonColor);
  FC:=ColorToRGB(FrameColor);
  Canvas.Brush.Color:=Warna;

  AwalR:=GetRValue(FC); AkhirR:=GetRValue(Warna);
  AwalG:=GetGValue(FC); AkhirG:=GetGValue(Warna);
  AwalB:=GetBValue(FC); AkhirB:=GetBValue(Warna);

  with Canvas do
  for n:=0 to t-1 do
  begin
    TempValue:=Sqrt(t*t-Sqr(t-n))/t;
    R:=AwalR+Round(TempValue*(AkhirR-AwalR));
    G:=AwalG+Round(TempValue*(AkhirG-AwalG));
    B:=AwalB+Round(TempValue*(AkhirB-AwalB));
    Brush.Color:=RGB(R, G, B);

    case Shape of
      shOval: FRgn:=CreateEllipticRgn(1+n, 1+n, Width-n, Height-n);
      shRectangle:
        begin
          Dia:=CornerRadius;
          if (Dia-n)>0 then
            FRgn:=CreateRoundRectRgn(1+n, 1+n, Width-n, Height-n, 2*(Dia-n), 2*(Dia-n))
          else FRgn:=CreateRectRgn(1+n, 1+n, Width-n-1, Height-n-1);
        end;
      shCapsule:
        begin
          if Width<Height then Dia:=Width div 2 else Dia:=Height div 2;
          if (Dia-n)>0 then
            FRgn:=CreateRoundRectRgn(1+n, 1+n, Width-n, Height-n, 2*(Dia-n), 2*(Dia-n))
          else FRgn:=CreateRectRgn(1+n, 1+n, Width-n-1, Height-n-1);
        end;
    end; //case
    try
      FrameRgn(Handle, FRgn, Brush.Handle, 1, 1);
    finally
      DeleteObject(FRgn);
    end;
  end;
end;

procedure TDsFancyButton.WriteCaption;
var Flags: Word;
    BtnL, BtnT, BtnR, BtnB: Integer;
    R, TR: TRect;
begin
  R:=ClientRect; TR:=ClientRect;
  Canvas.Font:=Self.Font;
  Canvas.Brush.Style:=bsClear;
  Flags:=DT_CENTER or DT_SINGLELINE;
  Canvas.Font:=Font;

  FTextColor:=Self.Font.Color;
  with canvas do
  begin
    BtnT:=TextT;
    BtnB:=BtnT+TextHeight(Caption);
    BtnL:=TextL;
    BtnR:=BtnL+TextWidth(Caption);
    TR:=Rect(BtnL, BtnT, BtnR, BtnB);
    R:=TR;

    if (TextStyle=txLowered) and not FIsDown then
    begin
      Font.Color:=clBtnHighLight;
      OffsetRect(TR, 1, 1);
      DrawText(Handle, PChar(Caption), Length(Caption), TR, Flags);
    end
    else if (TextStyle=txLowered) and FIsDown then
    begin
      Font.Color:=clBtnHighLight;
      OffsetRect(TR, +2, +2);
      DrawText(Handle, PChar(Caption), Length(Caption), TR, Flags);
    end
    else if (TextStyle=txRaised) and not FIsDown then
    begin
      Font.Color:=clBtnHighLight;
      OffsetRect(TR, -1, -1);
      DrawText(Handle, PChar(Caption), Length(Caption), TR, Flags);
    end
    else if (TextStyle=txRaised) and FIsDown then
    begin
      Font.Color:=clBtnHighLight;
      DrawText(Handle, PChar(Caption), Length(Caption), TR, Flags);
    end
    else if (TextStyle=txShadowed) and FIsDown then
    begin
      Font.Color:=clBtnShadow;
      OffsetRect(TR, 3, 3);
      DrawText(Handle, PChar(Caption), Length(Caption), TR, Flags);
    end
    else if (TextStyle=txShadowed) and not FIsDown then
    begin
      Font.Color:=clBtnShadow;
      OffsetRect(TR, 2, 2);
      DrawText(Handle, PChar(Caption), Length(Caption), TR, Flags);
    end;
    if Enabled then Font.Color:=FTextColor //Self.Font.Color
    else if (TextStyle=txShadowed) and not Enabled then
      Font.Color:=clBtnFace
    else Font.Color:=clBtnShadow;
    if FIsDown then OffsetRect(R, 1, 1);
    DrawText(Handle, PChar(Caption), Length(Caption), R, Flags);
  end;
end;

procedure TDsFancyButton.LayoutSetting;
var n: Integer;
    wg, hg, wt, ht: Integer;
    dist: Integer;
begin
  dist:=FGlyphSpace;
  wg:=FBmpGlyph.Width div FGlyphsNum;
  hg:=FBmpGlyph.Height;
  wt:=Canvas.TextWidth(Caption);
  ht:=Canvas.TextHeight(Caption);
  case GlyphLayout of
    lyLeft  : begin
                if Width>(wg+dist+wt) then GlyphL:=(Width-wg-dist-wt) div 2
                else GlyphL:=0;
                if Height>hg then GlyphT:=(Height-hg) div 2 else GlyphT:=0;
                TextL:=GlyphL+FBmpWidth+dist;
                TextT:=(Height-ht) div 2;
              end;
    lyRight : begin
                if Width>(wg+dist+wt) then n:=(Width-wg-dist-wt) div 2
                else n:=0;
                GlyphL:=Width-n-wg;
                if Height>hg then GlyphT:=(Height-hg) div 2 else GlyphT:=0;
                TextL:=GlyphL-wt-dist;
                TextT:=(Height-ht) div 2;
              end;
    lyTop   : begin
                if Width>wg then GlyphL:=(Width-wg) div 2 else GlyphL:=0;
                if Height>(hg+dist+ht) then GlyphT:=(Height-hg-dist-ht) div 2
                else GlyphT:=0;
                TextL:=(Width-wt) div 2;
                TextT:=GlyphT+FBmpGlyph.Height+dist;
              end;
    lyBottom: begin
                if Width>wg then GlyphL:=(Width-wg) div 2 else GlyphL:=0;
                if Height>(hg+dist+ht) then n:=(Height-hg-dist-ht) div 2
                else n:=0;
                GlyphT:=Height-n-hg;
                TextL:=(Width-wt) div 2;
                TextT:=GlyphT-dist-ht;
              end;
    lyCenter: begin
                if Width>wg then GlyphL:=(Width-wg) div 2 else GlyphL:=0;
                if Height>hg then GlyphT:=(Height-hg) div 2 else GlyphT:=0;
                TextL:=(Width-wt) div 2;
                TextT:=(Height-ht) div 2;
              end;
  end;
end;

procedure TDsFancyButton.DrawGlyph;
var c: Byte;
    R: TRect;
begin
  if FIsDown then
  begin
    R.Left:=GlyphL+1; R.Right:=GlyphL+1+FBmpWidth;
    R.Top:=GlyphT+1;  R.Bottom:=GlyphT+1+FBmpGlyph.Height;
  end
  else
  begin
    R.Left:=GlyphL; R.Right:=GlyphL+FBmpWidth;
    R.Top:=GlyphT;  R.Bottom:=GlyphT+FBmpGlyph.Height;
  end;

  { 0=enabled  1=disabled  2=pressed }
  if FIsDown and (FGlyphsNum=3) then c:=2
  else if Enabled then c:=0
  else if (not Enabled) and (FGlyphsNum>1) then c:=1
  else c:=0;

  Canvas.BrushCopy(R, FBmpGlyph,
                   Bounds(c*FBmpWidth, 0, FBmpWidth, FBmpGlyph.Height),
                   FTransparentColor);
  R:=ClientRect;
end;

procedure TDsFancyButton.SetBmpGlyph(ABmpGlyph: TBitmap);
begin
  FBmpGlyph.Assign(ABmpGlyph);
  if (FBmpGlyph<>nil) and (FTransparentColor=clNone) then
    FTransparentColor:=Glyph.Canvas.Pixels[0, Glyph.Height-1];
  Invalidate;
end;

procedure TDsFancyButton.SetBmpLayout(Value: TBmpLayout);
begin
  if Value<>FBmpLayout then
  begin
    FBmpLayout:=Value;
    Invalidate;
  end;
end;

procedure TDsFancyButton.SetGlyphsNum(Value: Byte);
begin
  FGlyphsNum:=Value;
  if FGlyphsNum<1 then FGlyphsNum:=1;
  if FGlyphsNum>3 then FGlyphsNum:=3;
  Invalidate;
end;

procedure TDsFancyButton.SetGlyphSpace(Value: Integer);
begin
  FGlyphSpace:=Value;
  if FGlyphSpace<1 then FGlyphSpace:=0;
  Invalidate;
end;

procedure TDsFancyButton.SetTransparentColor(NewColor: TColor);
begin
  if NewColor<>FTransparentColor then
  begin
    FTransparentColor:=NewColor;
    Invalidate;
  end;
end;

procedure TDsFancyButton.SetButtonColor(value: TColor);
begin
  if value<>FButtonColor then
    begin FButtonColor:=value; Invalidate; end;
end;

procedure TDsFancyButton.MouseDown(Button: TMouseButton;
          Shift: TShiftState; X, Y: Integer);
begin
  inherited MouseDown(Button, Shift, X, Y);
  if Button=mbRight then Exit;
  DrawButtonRegion;
  try
    if PtInRegion(MRgn, X, Y) then
      begin IsInRegion:=True; FIsDown:=True; Invalidate; end
    else IsInRegion:=False;
  finally
    DeleteObject(MRgn);
  end;
end;

procedure TDsFancyButton.MouseUp(Button: TMouseButton;
          Shift: TShiftState; X, Y: Integer);
begin
  inherited MouseUp(Button, Shift, X, Y);
  if Button=mbRight then Exit;
  if not IsInRegion then Exit;
  FIsDown:=False;
  Invalidate;
end;

procedure TDsFancyButton.SetShape(Value: TShape);
begin
  if Value<>FShape then
    begin FShape:=Value; Invalidate; end;
end;

procedure TDsFancyButton.SetTextStyle(Value: TTextStyle);
begin
  if Value<>FTextStyle then
    begin FTextStyle:=Value; Invalidate; end;
end;

procedure TDsFancyButton.SetFrameColor(Value: TColor);
begin
  if Value<>FFrameColor then
    begin FFrameColor:=Value; Invalidate; end;
end;

procedure TDsFancyButton.SetFrameWidth(Value: Integer);
var w: integer;
begin
  if Width<Height then w:=Width else w:=Height;
  if Value<>FFrameWidth then FFrameWidth:=Value;
  if FFrameWidth<4 then FFrameWidth:=4;
  if FFrameWidth>(w div 2) then FFrameWidth:=(w div 2);
  Invalidate;
end;

procedure TDsFancyButton.SetCornerRadius(Value: Integer);
var w: Integer;
begin
  if Width<Height then w:=Width else w:=Height;
  if Value<>FCornerRadius then FCornerRadius:=Value;
  if FCornerRadius<3 then FCornerRadius:=3;
  if FCornerRadius>w then FCornerRadius:=w;
  Invalidate;
end;

procedure TDsFancyButton.CMEnabledChanged(var Message: TMessage);
begin
  inherited;
  invalidate;
end;

procedure TDsFancyButton.CMTextChanged(var Message: TMessage);
begin
  Invalidate;
end;

procedure TDsFancyButton.CMDialogChar(var Message: TCMDialogChar);
begin
  with Message do
    if IsAccel(CharCode, Caption) and Enabled then
      begin Click; Result:=1; end
    else inherited;
end;

procedure TDsFancyButton.WMSize(var Message: TWMSize);
begin
  inherited;
  if Width>145 then Width:=145;
  if Height>145 then Height:=145;
end;

procedure TDsFancyButton.Click;
begin
  FIsDown:=False;
  Invalidate;
  inherited Click;
end;

procedure Register;
begin
  RegisterComponents('My Compo', [TDsFancyButton]);
end;

end.
